"""Module containing common plugin data definition."""

import queue
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from nxscli.logger import logger

if TYPE_CHECKING:
    from collections.abc import Callable

    from nxslib.dev import DeviceChannel
    from nxslib.nxscope import DNxscopeStream

    from nxscli.trigger import TriggerHandler

###############################################################################
# Class: PluginDataCb
###############################################################################


@dataclass
class PluginDataCb:
    """Plugin data callbacks."""

    stream_sub: "Callable[[int], queue.Queue[Any]]"
    stream_unsub: "Callable[[queue.Queue[Any]], None]"


###############################################################################
# Class: PluginQueueData
###############################################################################


class PluginQueueData:
    """The class used to handler stream queue data."""

    def __init__(
        self,
        que: queue.Queue[Any],
        channel: "DeviceChannel",
        trig: "TriggerHandler",
    ):
        """Initialize a queue data handler.

        :param que: queue subscribed to a given channel
        :param channel: instance of a channel
        :param dtc: trigger configuration
        """
        self._queue = que
        self._channel = channel
        self._trigger = trig

    def __str__(self) -> str:
        """Format string representation."""
        _str = (
            "PluginQueueData"
            + "(channel="
            + str(self._channel.data.chan)
            + ")"
        )
        return _str

    @property
    def queue(self) -> queue.Queue[Any]:
        """Get queue handler."""
        return self._queue

    @property
    def chan(self) -> int:
        """Get channel number."""
        return self._channel.data.chan

    @property
    def is_numerical(self) -> bool:
        """Return true if this channel is numerical."""
        return self._channel.data.is_numerical

    @property
    def vdim(self) -> int:
        """Return stream data dimension."""
        return self._channel.data.vdim

    @property
    def mlen(self) -> int:
        """Return stream metadata dimension."""
        return self._channel.data.mlen

    def queue_get(
        self, block: bool, timeout: float = 1.0
    ) -> list["DNxscopeStream"]:
        """Get data from a stream queue.

        :param block: blocking operation
        :param timeout: get data timeout
        """
        ret = []
        try:
            # get data from queue
            ret = self._queue.get(block=block, timeout=timeout)
        except queue.Empty:
            pass
        return self._trigger.data_triggered(ret)


###############################################################################
# Class: PluginData
###############################################################################


class PluginData:
    """A common plugin data handler."""

    def __init__(
        self,
        chanlist: list["DeviceChannel"],
        trig: list["TriggerHandler"],
        cb: PluginDataCb,
    ):
        """Initialize a plugin data handler.

        :param chanlist: a list with plugin channels
        :param cb: plugin callback to nxslib
        """
        logger.info("prepare data %s", str(chanlist))
        assert len(chanlist) == len(trig)

        self._chanlist = chanlist
        self._trig = trig
        self._cb = cb

        # queue handlers
        self._qdlist = self._qdlist_init()

    def __del__(self) -> None:
        """Deinitialize queue handlers."""
        try:
            self._queue_deinit()
        except AttributeError:
            pass

    def _qdlist_init(self) -> list[PluginQueueData]:
        ret = []
        for i, channel in enumerate(self._chanlist):
            # get queue with data
            que = self._cb.stream_sub(channel.data.chan)
            # initialize queue handler
            pdata = PluginQueueData(que, channel, self._trig[i])
            # add hanler to a list
            ret.append(pdata)
        return ret

    def _queue_deinit(self) -> None:
        """Deinitialize queue."""
        for i, pdata in enumerate(self._qdlist):
            self._cb.stream_unsub(pdata.queue)
            self._qdlist.pop(i)

        # clean up triggers
        # TODO: revisit where this beleong, here or in plugins ?
        for trig in self._trig:
            trig.cleanup()

    @property
    def qdlist(self) -> list[PluginQueueData]:
        """Get queue data handlers."""
        return self._qdlist



"""Module containint the Click environmet."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import click

if TYPE_CHECKING:
    from nxslib.nxscope import NxscopeHandler
    from nxslib.proto.parse import Parser

    from nxscli.phandler import PluginHandler
    from nxscli.trigger import DTriggerConfigReq


###############################################################################
# Class: DEnvironmentData
###############################################################################


@dataclass
class DEnvironmentData:
    """Environment data."""

    debug: bool = False
    waitenter: bool = False
    nxscope: "NxscopeHandler | None" = None
    parser: "Parser | None" = None
    interface: bool = False
    needchannels: bool = False
    channels: tuple[list[int], Any] | None = None
    phandler: "PluginHandler | None" = None
    triggers: dict[int, "DTriggerConfigReq"] | None = None


###############################################################################
# Class: Environment
###############################################################################


class Environment(DEnvironmentData):
    """A class with application environmet."""

    def __init__(self) -> None:
        """Initialize environmet."""
        super().__init__()


###############################################################################
# Decorator: pass_environment
###############################################################################


# custom environmet decorator
pass_environment = click.make_pass_decorator(Environment, ensure=True)



import pytest  # type: ignore

from nxscli.iplugin import EPluginType, IPlugin


class XTestPlugin1(IPlugin):
    def __init__(self, ptype):
        super().__init__(ptype)

    @property
    def stream(self) -> bool:
        return False

    def stop(self):
        pass

    def data_wait(self):
        return True

    def start(self, ploter, kwargs):
        return True

    def result(self):
        return


def test_nxscliplugin_init():
    # abstract class
    with pytest.raises(TypeError):
        _ = IPlugin()

    # valid plugin type
    p1 = XTestPlugin1(EPluginType.TEXT)

    assert p1.ptype == EPluginType.TEXT
    assert p1.handled is False
    p1.handled = True
    assert p1.handled is True
    assert p1.stream is False

    # connect dummy handler
    p1.connect_phandler(None)

    p1.stop()
    p1.data_wait()
    p1.start(None, None)
    p1.result()

    # at default plugins dont need to wait
    assert p1.wait_for_plugin() is True



"""Module containing Nxscli plugin interfaces."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from nxscli.phandler import PluginHandler

###############################################################################
# Enum: EPluginType
###############################################################################


class EPluginType(Enum):
    """Nxscli plugin types."""

    NONE = 1
    TEXT = 2
    STATIC = 3
    ANIMATION = 4
    FILE = 5


###############################################################################
# Data: DPluginDescription
###############################################################################


@dataclass
class DPluginDescription:
    """Plugin description."""

    name: str
    plugin: type["IPlugin"]


###############################################################################
# Class: IPlugin
###############################################################################


class IPlugin(ABC):
    """The Nxscli plugin common interface."""

    def __init__(self, ptype: EPluginType) -> None:
        """Initialize a Nxslib plugin.

        :param ptype: plugin type
        """
        self._ptype = ptype
        self._handled = False
        self._phandler: "PluginHandler"

    @property
    def ptype(self) -> EPluginType:
        """Get plugin type."""
        return self._ptype

    @property
    def handled(self) -> bool:
        """Get handled flag."""
        return self._handled

    @handled.setter
    def handled(self, val: bool) -> None:
        """Set handled flag.

        :param val: plugin handled state
        """
        self._handled = val

    def wait_for_plugin(self) -> bool:
        """Return True if plugin is dont't need to wait."""
        return True

    @property
    @abstractmethod
    def stream(self) -> bool:
        """Return True if this plugin needs stream."""

    def connect_phandler(self, phandler: "PluginHandler") -> None:
        """Connect phandler.

        :param phandler: plugin handler instance
        """
        self._phandler = phandler

    @abstractmethod
    def stop(self) -> None:
        """Interface method."""

    @abstractmethod
    def start(self, kwargs: Any) -> bool:
        """Interface method.

        :param kwargs: plugin specific arguments
        """

    @abstractmethod
    def data_wait(self, timeout: float = 0.0) -> bool:
        """Return True if data are ready.

        :param timeout: data wait timeout
        """

    @abstractmethod
    def result(self) -> Any:
        """Interface method."""


###############################################################################
# Class: IPluginNone
###############################################################################


class IPluginNone(IPlugin):
    """None-type plugin."""

    def __init__(self) -> None:
        """Initialize none-type plugin."""
        super().__init__(EPluginType.NONE)


###############################################################################
# Class: IPluginText
###############################################################################


class IPluginText(IPlugin):
    """Text-type plugin."""

    def __init__(self) -> None:
        """Initialize text-type plugin."""
        super().__init__(EPluginType.TEXT)


###############################################################################
# Class: IPluginPlotStatic
###############################################################################


class IPluginPlotStatic(IPlugin):
    """Static-plot plugin."""

    def __init__(self) -> None:
        """Initialize static-plot-type plugin."""
        super().__init__(EPluginType.STATIC)


###############################################################################
# Class: IPluginPlotDynamic
###############################################################################


class IPluginPlotDynamic(IPlugin):
    """Dynamic-plot plugin."""

    def __init__(self) -> None:
        """Initialize dynamic-plot-type plugin."""
        super().__init__(EPluginType.ANIMATION)


###############################################################################
# Class: IPluginFile
###############################################################################


class IPluginFile(IPlugin):
    """File-type plugin."""

    def __init__(self) -> None:
        """Initialize file-type plugin."""
        super().__init__(EPluginType.FILE)



import queue

import pytest  # type: ignore
from nxslib.dev import DeviceChannel

from nxscli.idata import PluginData, PluginDataCb, PluginQueueData
from nxscli.trigger import DTriggerConfig, ETriggerType, TriggerHandler

g_queue: queue.Queue[list] = queue.Queue()


def dummy_stream_sub(ch):
    return g_queue


def dummy_stream_unsub(q):
    assert q == g_queue


def test_pluginqueuedata():
    q = queue.Queue()
    chan = DeviceChannel(0, 2, 2, "chan0")
    dtc = DTriggerConfig(ETriggerType.ALWAYS_OFF)
    trig = TriggerHandler(0, dtc)
    qdata = PluginQueueData(q, chan, trig)

    assert isinstance(str(qdata), str)
    assert qdata.queue == q
    assert qdata.chan == chan.data.chan
    assert qdata.is_numerical == chan.data.is_numerical
    assert qdata.vdim == chan.data.vdim
    assert qdata.mlen == chan.data.mlen

    # no data on queue
    ret = qdata.queue_get(block=False)
    assert ret == []
    ret = qdata.queue_get(block=True, timeout=0.1)
    assert ret == []

    TriggerHandler.cls_cleanup()


def test_nxsclipdata_init():
    channels = [DeviceChannel(0, 1, 2, "chan0")]
    dtc = DTriggerConfig(ETriggerType.ALWAYS_OFF)
    trig = [TriggerHandler(0, dtc)]
    cb = PluginDataCb(dummy_stream_sub, dummy_stream_unsub)

    with pytest.raises(AssertionError):
        gdata = PluginData(channels, [], cb)

    gdata = PluginData(channels, trig, cb)

    assert gdata.qdlist[0].queue is g_queue
    assert gdata.qdlist[0].chan == 0
    assert gdata.qdlist[0].is_numerical is False
    assert gdata.qdlist[0].vdim == 2
    assert gdata.qdlist[0].mlen == 0

    TriggerHandler.cls_cleanup()



"""Module containing printer plugin."""

import queue
from threading import Event
from typing import TYPE_CHECKING, Any

from nxscli.idata import PluginData, PluginQueueData
from nxscli.iplugin import IPluginText
from nxscli.logger import logger
from nxscli.pluginthr import PluginThread

if TYPE_CHECKING:
    from nxslib.nxscope import DNxscopeStream


###############################################################################
# Class: PluginPrinter
###############################################################################


class PluginPrinter(PluginThread, IPluginText):
    """Dummy plugin that print captured data."""

    def __init__(self) -> None:
        """Intiialize a printer plugin."""
        IPluginText.__init__(self)
        PluginThread.__init__(self)

        self._q: queue.Queue[Any] = queue.Queue()
        self._done = Event()
        self._ret_len = 0
        self._meta_string = False

        self._data: "PluginData"

    def _init(self) -> None:
        assert self._phandler

    def _final(self) -> None:
        logger.info("printer DONE")

    def _handle_samples(
        self, data: list["DNxscopeStream"], pdata: "PluginQueueData", j: int
    ) -> None:
        for sample in data:
            if self._datalen[j] < self._samples:
                d: dict[str, Any] = dict()
                d["chan"] = self._data.qdlist[j].chan
                d["data"] = sample.data
                if self._meta_string:
                    d["meta"] = bytes(list(sample.meta)).decode()
                else:
                    d["meta"] = sample.meta
                self._q.put(d)
                self._datalen[j] += 1

    @property
    def handled(self) -> bool:
        """Get handled flag."""
        return self._handled

    @handled.setter
    def handled(self, val: bool) -> None:
        """Overload method common method.

        :param val: plugin handled state
        """
        if self._done.is_set():
            self._handled = val
        else:
            # force not handled
            self._handled = False

    def data_wait(self, timeout: float = 0.0) -> bool:
        """Return True if data are ready.

        :param timeout: data wait timeout
        """
        return True

    def start(self, kwargs: Any) -> bool:
        """Start printer plugin.

        :param kwargs: implementation specific arguments
        """
        assert self._phandler

        logger.info("start capture %s", str(kwargs))

        self._samples = kwargs["samples"]
        self._meta_string = kwargs["metastr"]
        self._nostop = kwargs["nostop"]

        chanlist = self._phandler.chanlist_plugin(kwargs["channels"])
        trig = self._phandler.triggers_plugin(chanlist, kwargs["trig"])

        cb = self._phandler.cb_get()
        self._data = PluginData(chanlist, trig, cb)

        if not self._data.qdlist:  # pragma: no cover
            return False

        self.thread_start(self._data)

        return True

    def result(self) -> str:
        """Get printer plugin result."""
        try:
            samples = self._q.get(block=True, timeout=1.0)
            self._ret_len += 1

        except queue.Empty:
            return ""

        if self._ret_len >= self._samples:
            self._done.set()

        s = str(self._ret_len) + ": " + str(samples)
        return s



"""Module containing dummy capture plugin."""

from typing import TYPE_CHECKING, Any

from nxscli.idata import PluginData, PluginQueueData
from nxscli.iplugin import IPluginNone
from nxscli.logger import logger
from nxscli.pluginthr import PluginThread

if TYPE_CHECKING:
    from nxslib.nxscope import DNxscopeStream


###############################################################################
# Class: PluginNone
###############################################################################


class PluginNone(PluginThread, IPluginNone):
    """Dummy plugin that do nothing with captured data."""

    def __init__(self) -> None:
        """Intiialize a none plugin."""
        IPluginNone.__init__(self)
        PluginThread.__init__(self)

        self._data: "PluginData"

    def _init(self) -> None:
        assert self._phandler

    def _final(self) -> None:
        logger.info("None DONE")

    def _handle_samples(
        self, data: list["DNxscopeStream"], pdata: "PluginQueueData", j: int
    ) -> None:
        for _ in data:
            # get data len
            self._datalen[j] += 1

    def start(self, kwargs: Any) -> bool:
        """Start none plugin.

        :param kwargs: implementation specific arguments
        """
        assert self._phandler

        logger.info("start capture %s", str(kwargs))

        self._samples = kwargs["samples"]
        self._nostop = kwargs["nostop"]

        chanlist = self._phandler.chanlist_plugin(kwargs["channels"])
        trig = self._phandler.triggers_plugin(chanlist, kwargs["trig"])

        cb = self._phandler.cb_get()
        self._data = PluginData(chanlist, trig, cb)

        if not self._data.qdlist:  # pragma: no cover
            return False

        self.thread_start(self._data)

        return True

    def result(self) -> None:
        """Get none plugin result."""
        return  # pragma: no cover



"""Module containint the dummy interface command for CLI."""

import click
from nxslib.intf.dummy import DummyDev
from nxslib.nxscope import NxscopeHandler

from nxscli.cli.environment import Environment, pass_environment

###############################################################################
# Command: cmd_dummy
###############################################################################


@click.group(name="dummy", chain=True)
@click.option("--writepadding", default=0)
@click.option(
    "--streamsleep", type=float, default=0.001, help="dummy dev parameter"
)
@click.option(
    "--samplesnum", type=int, default=100, help="dummy dev parameter"
)
@pass_environment
def cmd_dummy(
    ctx: Environment, writepadding: int, streamsleep: float, samplesnum: int
) -> bool:
    """[interface] Connect with a simulated NxScope devicve.

    \b
    Channels data:
      chan0 - vdim = 1, random()
      chan1 - vdim = 1, saw wave
      chan2 - vdim = 1, triangle wave
      chan3 - vdim = 2, random()
      chan4 - vdim = 3, random()
      chan5 - vdim = 3, static vector = [1.0, 0.0, -1.0]
      chan6 - vdim = 1, 'hello' string
      chan7 - vdim = 3, static vector = [1.0, 0.0, -1.0], meta = 1B int
      chan8 - vdim = 0, meta = 'hello string', mlen = 16
      chan9 - vdim = 3, 3-phase sine wave
    """  # noqa: D301
    intf = DummyDev(
        rxpadding=writepadding,
        stream_sleep=streamsleep,
        stream_snum=samplesnum,
    )

    # initialize nxslib communication handler
    assert ctx.parser
    ctx.nxscope = NxscopeHandler(intf, ctx.parser)

    ctx.interface = True

    return True



"""Module containt the nxscli handler implementation."""

from typing import TYPE_CHECKING, Any

from nxscli.idata import PluginDataCb
from nxscli.logger import logger
from nxscli.trigger import DTriggerConfigReq, TriggerHandler, trigger_from_req

if TYPE_CHECKING:
    from nxslib.dev import Device, DeviceChannel
    from nxslib.nxscope import NxscopeHandler

    from nxscli.iplugin import DPluginDescription, IPlugin


###############################################################################
# Class: PluginHandler
###############################################################################


class PluginHandler:
    """A class implementing a plugins handler."""

    def __init__(self, plugins: list["DPluginDescription"] | None = None):
        """Initialize a plugin handler.

        :param plugins: a list with plugins
        """
        self._nxs: "NxscopeHandler" | None = None
        self._plugins = {}

        if plugins:
            for cls in plugins:
                self._plugins[cls.name] = cls.plugin

        self._enabled: list[tuple[int, type["IPlugin"], Any]] = []
        self._started: list[tuple["IPlugin", Any]] = []
        self._triggers: dict[int, DTriggerConfigReq] = {}

        self._chanlist: list["DeviceChannel"] = []

        # stream flags
        self._stream = False

        self._cleanup_done = False

    def __del__(self) -> None:
        """Raise assertion if not cleaned."""
        if not self._cleanup_done:
            raise AssertionError("PluginHandler not cleaned")

    def _chanlist_gen(self, channels: list[int]) -> list["DeviceChannel"]:
        assert self._nxs
        assert self.dev
        # convert special keys for all channels
        if channels and channels[0] == -1:  # pragma: no cover
            chanlist = list(range(self.dev.data.chmax))
        else:
            assert all(isinstance(x, int) for x in channels)
            chanlist = channels

        # get channels data
        ret = []
        for chan in chanlist:
            channel = self._nxs.dev_channel_get(chan)
            assert channel
            ret.append(channel)

        return ret

    def _chanlist_enable(self) -> None:
        assert self._nxs
        for channel in self._chanlist:
            # ignore not valid channels
            if not channel.data.is_valid:  # pragma: no cover
                logger.info(
                    "NOTE: channel %d not valid - ignore", channel.data.chan
                )
                continue

            # enable channel
            self._nxs.ch_enable(channel.data.chan)

    def _chanlist_div(self, div: int | list[int]) -> None:
        assert self._nxs
        if isinstance(div, int):
            for channel in self._chanlist:
                self._nxs.ch_divider(channel.data.chan, div)
        else:
            # divider list configuration must cover all configured channels
            assert len(div) == len(self._chanlist)
            for i, channel in enumerate(self._chanlist):
                self._nxs.ch_divider(channel.data.chan, div[i])

    @property
    def chanlist(self) -> list["DeviceChannel"]:
        """Get configured channels list."""
        return self._chanlist

    @property
    def names(self) -> list[str]:
        """Get plugins names."""
        return list(self._plugins.keys())

    @property
    def plugins(self) -> dict[str, type["IPlugin"]]:
        """Get loaded plugins."""
        return self._plugins

    @property
    def dev(self) -> "Device | None":
        """Get Nxscope device info."""
        assert self._nxs
        return self._nxs.dev

    @property
    def stream(self) -> bool:
        """Get stream flag."""
        return self._stream

    @property
    def enabled(self) -> list[tuple[int, type["IPlugin"], Any]]:
        """Get enabled plugins."""
        return self._enabled

    def cleanup(self) -> None:
        """Clean up - must be called after instance use."""
        # disconnect nxscope if connected
        self.nxscope_disconnect()
        # clean up triggers
        TriggerHandler.cls_cleanup()
        self._cleanup_done = True

    def cb_get(self) -> PluginDataCb:
        """Get callbacks for plugins."""
        assert self._nxs
        return PluginDataCb(self._nxs.stream_sub, self._nxs.stream_unsub)

    def stream_start(self) -> None:
        """Start stream."""
        assert self._nxs
        self._nxs.stream_start()

    def stream_stop(self) -> None:
        """Stop stream."""
        assert self._nxs
        self._nxs.stream_stop()

    def nxscope_disconnect(self) -> None:
        """Disconnect from NxScope device."""
        if self._nxs:
            logger.info("disconnecting from nxs device...")
            # connect nxscope device
            self._nxs.disconnect()
            logger.info("disconnected!")
            self._nxs = None

    def nxscope_connect(self, nxs: "NxscopeHandler") -> None:
        """Connect Nxslib instance.

        :param nxs: Nxscope handler
        """
        self._nxs = nxs
        logger.info("connecting to nxs device...")
        # connect nxscope device
        self._nxs.connect()
        logger.info("connected!")

    def plugin_add(self, cls: tuple[str, type["IPlugin"]]) -> None:
        """Add plugin.

        :param cls: tuple with plugin data
        """
        self._plugins[cls[0]] = cls[1]

    def plugin_get(self, name: str) -> type["IPlugin"]:
        """Get plugin by name.

        :param name: plugin name
        """
        return self._plugins[name]

    def enable(self, name: str, **kwargs: Any) -> int:
        """Enable plugin.

        :param name: plugin name
        :param kwargs: implementation specific arguments
        """
        pid = len(self._enabled)
        plugin = (pid, self._plugins[name], kwargs)
        logger.info("enable %s", str(plugin))
        self._enabled.append(plugin)
        return pid

    def disable(self, pid: int) -> None:
        """Disable plugin with a given ID.

        :param pid: disable plugin by ID
        """
        found = False
        i = 0
        for x in self._enabled:
            if pid == x[0]:
                found = True
                break
            i += 1
        if found:
            logger.info("disable %d %s", pid, str(self._enabled[i]))
            self._enabled.pop(i)
        else:
            raise AttributeError

    def start(self) -> None:
        """Start all enabled plugins."""
        for plg in self._enabled:
            pid, cls, args = plg

            # create instance
            plugin = cls()  # type: ignore

            # we need data stream
            if plugin.stream is True:
                self._stream = True

            plugin.connect_phandler(self)
            if not plugin.start(args):  # pragma: no cover
                logger.error("failed to start plugin %s", str(plugin))
            else:
                self._started.append((plugin, args))
            logger.info("started %s", str(plugin))

        # start stream if needed
        if self.stream:
            self.stream_start()

    def stop(self) -> None:
        """Stop stream and all started plugins."""
        if self._stream is True:
            self.stream_stop()

        for plg, _ in self._started:
            plg.stop()

    def ready(self) -> list["IPlugin"]:
        """Wait for results from enabled plugins."""
        ret = []
        for plg in self._started:
            plugin, _ = plg
            # blocking
            if plugin.data_wait():
                ret.append(plugin)

        return ret

    def poll(self) -> list["IPlugin"] | None:
        """Pool for results from enabled plugins."""
        nothandled = 0
        for plg in self._started:
            plugin, _ = plg
            if not plugin.handled:
                nothandled += 1
        if not nothandled:
            return None

        ret = []
        for plg in self._started:
            plugin, _ = plg
            if not plugin.handled:
                if plugin.data_wait(0.1):
                    ret.append(plugin)
                    plugin.handled = True
            else:  # pragma: no cover
                pass

        return ret

    def wait_for_plugins(self) -> None:
        """Wait for plugins."""
        while True:
            done = True
            for plg in self._started:
                plugin, _ = plg
                if not plugin.wait_for_plugin():
                    done = False
            # exit from loop if done
            if done:
                break

    def trigger_get(
        self, chid: int, src: dict[int, DTriggerConfigReq] | None = None
    ) -> DTriggerConfigReq:
        """Get trigger for a given channel.

        :param chid: channel ID
        :param src: trigger configuration source
        """
        # get data from plugin private dict or global dict
        if src:
            array = src  # pragma: no cover
        else:
            array = self._triggers

        # get trigger configuration from a given source
        try:
            trg = array[chid]
        except KeyError:
            # check for global configuration key
            globkey = -1
            if globkey in array:
                trg = array[globkey]
            else:
                # default on
                trg = DTriggerConfigReq("on", None)
        return trg

    def chanlist_plugin(self, channels: list[int]) -> list["DeviceChannel"]:
        """Prepare channels list for a plugin.

        :param chanlist: a list with plugin channels
        """
        chanlist = []
        if channels and channels[0] != -1:
            # plugin specific channels configuration
            for chan in self.chanlist:  # pragma: no cover
                if chan.data.chan in channels:
                    chanlist.append(chan)
                else:  # pragma: no cover
                    pass
        else:
            chanlist = self.chanlist

        return chanlist

    def channels_configure(
        self,
        channels: list[int],
        div: int | list[int] = 0,
        writenow: bool = False,
    ) -> None:
        """Configure channels.

        The effects of this method are buffered and will
        be applied to the device just before the stream starts
        or can be forced to write with writenow flag.

        :param chans: a list with channels IDs
        :param div: a list with divider values
        :param writenow: write channels configuration now
        """
        assert self._nxs

        logger.info("configure channels = %s divider = %d", str(channels), div)

        self._chanlist = self._chanlist_gen(channels)
        if not self._chanlist:
            return

        # default channels configuration
        self._nxs.channels_default_cfg()

        # enable channels
        self._chanlist_enable()

        # set divider for channels
        self._chanlist_div(div)

        if writenow:
            # write channels configuration
            self._nxs.channels_write()

    def triggers_configure(
        self, triggers: dict[int, DTriggerConfigReq]
    ) -> None:
        """Configure triggers.

        :param triggers: dict with triggers configuration
        """
        self._triggers = triggers

    def triggers_plugin(
        self,
        chanlist: list["DeviceChannel"],
        triggers: dict[int, DTriggerConfigReq] | None,
    ) -> list[TriggerHandler]:
        """Prepare triggers list for a plugin.

        :param chanlist: a list with plugin channels
        :param triggers: a list with plugin triggers
        """
        trgs: list[tuple[int, DTriggerConfigReq]] = []
        if triggers:
            # plugin specific triggers
            for chan in chanlist:  # pragma: no cover
                tcfg = self.trigger_get(chan.data.chan, triggers)
                trgs.append((chan.data.chan, tcfg))
        else:
            # global configured triggers
            for chan in chanlist:
                tcfg = self.trigger_get(chan.data.chan)
                trgs.append((chan.data.chan, tcfg))

        ret = []
        for item in trgs:
            # get trigger configuration
            dtc = trigger_from_req(item[1])
            # get trigger
            trig = TriggerHandler(item[0], dtc)
            ret.append(trig)

        return ret



"""Default commands."""

from typing import TYPE_CHECKING

from nxscli.commands.cmd_csv import cmd_pcsv
from nxscli.commands.cmd_devinfo import cmd_pdevinfo
from nxscli.commands.cmd_none import cmd_pnone
from nxscli.commands.cmd_printer import cmd_printer
from nxscli.commands.config.cmd_chan import cmd_chan
from nxscli.commands.config.cmd_trig import cmd_trig

if TYPE_CHECKING:
    import click

commands_list: list["click.Command"] = [
    cmd_chan,
    cmd_trig,
    cmd_pdevinfo,
    cmd_pcsv,
    cmd_pnone,
    cmd_printer,
]



"""Module containing devinfo plugin."""

import pprint
from typing import Any

from nxscli.iplugin import IPluginText

###############################################################################
# Class: PluginDevinfo
###############################################################################


class PluginDevinfo(IPluginText):
    """Plugin that shows device information."""

    def __init__(self) -> None:
        """Initialize devinfo plugin."""
        super().__init__()
        self._return = None

    @property
    def stream(self) -> bool:
        """Return True if this plugin needs stream."""
        return False

    def stop(self) -> None:
        """Stop devinfo plugin."""

    def data_wait(self, timeout: float = 0.0) -> bool:
        """Return True if data are ready.

        :param timeout: not used
        """
        return True

    def start(self, _: Any) -> bool:
        """Start devinfo plugin."""
        assert self._phandler
        assert self._phandler.dev

        dev = self._phandler.dev

        ret: Any = {}
        ret["cmn"] = {}
        ret["cmn"]["chmax"] = dev.data.chmax
        ret["cmn"]["flags"] = dev.data.flags
        ret["cmn"]["rxpadding"] = dev.data.rxpadding

        tmp = []
        for chid in range(dev.data.chmax):
            chinfo = dev.channel_get(chid)
            assert chinfo
            chan: Any = {}
            chan["chan"] = chinfo.data.chan
            chan["type"] = chinfo.data._type
            chan["vdim"] = chinfo.data.vdim
            chan["name"] = chinfo.data.name

            tmp.append(chan)

        ret["channels"] = tmp

        self._return = ret

        return True

    def result(self) -> str:
        """Get devinfo plugin result."""
        assert self._return
        s = "\nDevice common:\n"
        s += pprint.pformat(self._return["cmn"])
        s += "\nDevice channels:\n"
        s += pprint.pformat(self._return["channels"])
        s += "\n"
        return s



"""Default interfaces."""

from typing import TYPE_CHECKING

from nxscli.commands.interface.cmd_dummy import cmd_dummy
from nxscli.commands.interface.cmd_serial import cmd_serial

if TYPE_CHECKING:
    import click

interfaces_list: list["click.Group"] = [cmd_serial, cmd_dummy]



"""Module containint the serial interface command for CLI."""

import click
from nxslib.intf.serial import SerialDevice
from nxslib.nxscope import NxscopeHandler

from nxscli.cli.environment import Environment, pass_environment

###############################################################################
# Command: cmd_serial
###############################################################################


@click.group(name="serial", chain=True)
@click.argument("path", type=click.Path(resolve_path=False), required=True)
@click.option("--baud", default=115200)
@click.option("--writepadding", default=0)
@pass_environment
def cmd_serial(
    ctx: Environment, path: str, baud: int, writepadding: bool
) -> bool:  # pragma: no cover
    """[interface] Connect with a serial port NxScope devie."""
    intf = SerialDevice(path, baud=baud)
    intf.write_padding = writepadding

    # initialize nxslib communication handler
    assert ctx.parser
    ctx.nxscope = NxscopeHandler(intf, ctx.parser)

    ctx.interface = True

    return True



from threading import Lock

import pytest  # type: ignore
from nxslib.nxscope import DNxscopeStream

from nxscli.trigger import (
    DTriggerConfig,
    DTriggerConfigReq,
    ETriggerType,
    TriggerHandler,
    trigger_from_req,
)

# we want to run TriggerHandler tests without concurency
global_lock = Lock()


def test_triggerfromstr():
    req = DTriggerConfigReq("off", None)
    x = trigger_from_req(req)
    assert x.ttype == ETriggerType.ALWAYS_OFF
    assert x.srcchan is None
    assert x.hoffset == 0
    assert x.level is None

    req = DTriggerConfigReq("on", None)
    x = trigger_from_req(req)
    assert x.ttype == ETriggerType.ALWAYS_ON
    assert x.srcchan is None
    assert x.hoffset == 0
    assert x.level is None

    req = DTriggerConfigReq("er", None, params=["100", "10"])
    x = trigger_from_req(req)
    assert x.ttype == ETriggerType.EDGE_RISING
    assert x.hoffset == 100
    assert x.level == 10

    req = DTriggerConfigReq("ef", None, params=["200", "20"])
    x = trigger_from_req(req)
    assert x.ttype == ETriggerType.EDGE_FALLING
    assert x.srcchan is None
    assert x.hoffset == 200
    assert x.level == 20

    req = DTriggerConfigReq("ef", 1, params=["200", "20"])
    x = trigger_from_req(req)
    assert x.ttype == ETriggerType.EDGE_FALLING
    assert x.srcchan == 1
    assert x.hoffset == 200
    assert x.level == 20


# initialization logic
def test_triggerhandle_init():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        dtc0 = DTriggerConfig(ETriggerType.ALWAYS_OFF)
        th0 = TriggerHandler(0, dtc0)

        assert th0.chan == 0
        assert th0._src is None
        assert th0._cross == []
        assert len(th0._instances) == 1

        # we can register many trigger handlers for the same channel
        dtc0_1 = DTriggerConfig(ETriggerType.ALWAYS_ON)
        th0_1 = TriggerHandler(0, dtc0_1)

        assert th0_1.chan == 0
        assert th0_1._src is None
        assert th0_1._cross == []

        assert len(th0._instances) == 2
        assert len(th0_1._instances) == 2

        # no source channel registerd
        dtcx = DTriggerConfig(ETriggerType.ALWAYS_OFF, srcchan=2)
        thfail = TriggerHandler(1, dtcx)

        # one is waiting instnaces
        assert len(TriggerHandler._wait_for_src) == 1

        # try get data but we don't have required src channel
        with pytest.raises(AssertionError):
            din = [
                DNxscopeStream((0,), ()),
                DNxscopeStream((0,), ()),
                DNxscopeStream((0,), ()),
            ]
            _ = thfail.data_triggered(din)

        assert th0.chan == 0
        assert th0._src is None
        assert th0._cross == []

        assert len(th0._instances) == 3
        assert len(th0_1._instances) == 3

        # valid cross-channel source trigger
        dtc1 = DTriggerConfig(ETriggerType.ALWAYS_OFF, srcchan=0)
        th1 = TriggerHandler(1, dtc1)

        assert th0.chan == 0
        assert th0._src is None

        assert th0_1.chan == 0
        assert th0_1._src is None

        assert th1.chan == 1
        assert th1._src is not None
        assert th1 in th1._src._cross
        assert th1._cross == []

        assert len(th0._instances) == 4
        assert len(th0_1._instances) == 4
        assert len(th1._instances) == 4

        # another valid cross-channel source trigger
        dtc2 = DTriggerConfig(ETriggerType.ALWAYS_OFF, srcchan=0)
        th2 = TriggerHandler(2, dtc2)

        # another valid cross-channel source trigger
        dtc3 = DTriggerConfig(ETriggerType.ALWAYS_OFF, srcchan=0)
        th3 = TriggerHandler(3, dtc3)

        assert len(th0._instances) == 6
        assert len(th0_1._instances) == 6
        assert len(th1._instances) == 6
        assert len(th2._instances) == 6
        assert len(th3._instances) == 6

        assert th2._src is not None
        assert th2 in th2._src._cross

        assert th3._src is not None
        assert th3 in th3._src._cross

        # clean up
        TriggerHandler.cls_cleanup()


# check if clean up works
def test_triggerhandle_init2():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        dtc0 = DTriggerConfig(ETriggerType.ALWAYS_OFF)
        th0 = TriggerHandler(0, dtc0)

        # there should be no references to the previous test
        assert len(th0._instances) == 1

        # clean up
        TriggerHandler.cls_cleanup()


# cross trigger but no source channel during trigger registration
def test_triggerhandle_init3():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        dtc1 = DTriggerConfig(
            ETriggerType.EDGE_RISING, srcchan=0, hoffset=0, level=0
        )
        th1 = TriggerHandler(1, dtc1)

        assert th1.chan == 1
        assert th1._src is None  # no reference now
        assert th1._cross == []
        assert len(th1._instances) == 1

        # one is waiting instnaces
        assert len(TriggerHandler._wait_for_src) == 1

        # try get data but we don't have required src channel yet
        with pytest.raises(AssertionError):
            din = [
                DNxscopeStream((0,), ()),
                DNxscopeStream((0,), ()),
                DNxscopeStream((0,), ()),
            ]
            _ = th1.data_triggered(din)

        dtc0 = DTriggerConfig(ETriggerType.ALWAYS_OFF)
        th0 = TriggerHandler(0, dtc0)

        assert th0.chan == 0
        assert th0._src is None
        assert th1 in th0._cross
        assert len(th0._instances) == 2

        # no waiting instnaces
        assert len(TriggerHandler._wait_for_src) == 0

        # try get data, now should not fail
        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        _ = th1.data_triggered(din)

        # wait for source channel and celan up without registered source
        dtc3 = DTriggerConfig(
            ETriggerType.EDGE_RISING, srcchan=5, hoffset=0, level=0
        )
        th3 = TriggerHandler(3, dtc3)

        assert th3._src is None
        assert len(TriggerHandler._wait_for_src) == 1

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_alwaysoff():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # always off
        dtc = DTriggerConfig(ETriggerType.ALWAYS_OFF)
        th = TriggerHandler(0, dtc)
        assert len(th._instances) == 1
        for _ in range(100):
            din = [
                DNxscopeStream((1,), ()),
                DNxscopeStream((2,), ()),
                DNxscopeStream((3,), ()),
            ]
            dout = th.data_triggered(din)
            assert dout == []

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_alwayson():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # always on
        dtc = DTriggerConfig(ETriggerType.ALWAYS_ON)
        th = TriggerHandler(0, dtc)
        assert len(th._instances) == 1
        for _ in range(100):
            din = [
                DNxscopeStream((1,), ()),
                DNxscopeStream((2,), ()),
                DNxscopeStream((3,), ()),
            ]
            dout = th.data_triggered(din)
            assert dout == [
                DNxscopeStream((1,), ()),
                DNxscopeStream((2,), ()),
                DNxscopeStream((3,), ()),
            ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_edgerising1():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # rising edge on 0
        hoffset = 0
        level = 0
        dtc = DTriggerConfig(
            ETriggerType.EDGE_RISING, hoffset=hoffset, level=level
        )
        th = TriggerHandler(0, dtc)

        assert len(th._instances) == 1

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
            DNxscopeStream((-3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((-3,), ()),
            DNxscopeStream((-3,), ()),
            DNxscopeStream((-3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        # triggered - rising edge on 0
        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]

        din = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
        ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_edgerising2():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # rising edge on 5
        hoffset = 0
        level = 5
        dtc = DTriggerConfig(
            ETriggerType.EDGE_RISING, hoffset=hoffset, level=level
        )
        th = TriggerHandler(0, dtc)

        assert len(th._instances) == 1

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
            DNxscopeStream((-3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((-4,), ()),
            DNxscopeStream((-3,), ()),
            DNxscopeStream((-2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        # triggered
        din = [
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
            DNxscopeStream((6,), ()),
            DNxscopeStream((7,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((5,), ()),
            DNxscopeStream((6,), ()),
            DNxscopeStream((7,), ()),
        ]

        din = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
        ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_edgefalling1():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # falling edge on 0
        hoffset = 0
        level = 0
        dtc = DTriggerConfig(
            ETriggerType.EDGE_FALLING, hoffset=hoffset, level=level
        )
        th = TriggerHandler(0, dtc)

        assert len(th._instances) == 1

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        # triggered
        din = [
            DNxscopeStream((2,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [DNxscopeStream((0,), ()), DNxscopeStream((-1,), ())]

        din = [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
            DNxscopeStream((-3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
            DNxscopeStream((-3,), ()),
        ]

        din = [
            DNxscopeStream((2,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((2,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
        ]

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]

        din = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
            DNxscopeStream((3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
            DNxscopeStream((3,), ()),
        ]

        din = [
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
            DNxscopeStream((6,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
            DNxscopeStream((6,), ()),
        ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_edgefalling2():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # falling edge on -5
        hoffset = 0
        level = -5
        dtc = DTriggerConfig(
            ETriggerType.EDGE_FALLING, hoffset=hoffset, level=level
        )
        th = TriggerHandler(0, dtc)

        assert len(th._instances) == 1

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((2,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        din = [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-2,), ()),
            DNxscopeStream((-3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == []

        # triggered
        din = [
            DNxscopeStream((-4,), ()),
            DNxscopeStream((-5,), ()),
            DNxscopeStream((-6,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [DNxscopeStream((-5,), ()), DNxscopeStream((-6,), ())]

        din = [
            DNxscopeStream((2,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((2,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
        ]

        din = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]

        din = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
            DNxscopeStream((3,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
            DNxscopeStream((3,), ()),
        ]

        din = [
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
            DNxscopeStream((6,), ()),
        ]
        dout = th.data_triggered(din)
        assert dout == [
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
            DNxscopeStream((6,), ()),
        ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_chanxtochany_nohoffset():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # chan0 - always off
        dtc0 = DTriggerConfig(ETriggerType.ALWAYS_OFF)
        th0 = TriggerHandler(0, dtc0)

        # chan1 - trigger on chan0 rising endge 4
        hoffset = 0
        level = 4
        srcchan = 0
        dtc1 = DTriggerConfig(
            ETriggerType.EDGE_RISING,
            srcchan=srcchan,
            hoffset=hoffset,
            level=level,
        )
        th1 = TriggerHandler(1, dtc1)

        assert len(th0._instances) == 2
        assert len(th1._instances) == 2

        din0 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []

        din0 = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []

        din0 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((4,), ()),
            DNxscopeStream((3,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []

        # th1 triggered from now - but th0 is always off
        din0 = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((10,), ()),
            DNxscopeStream((11,), ()),
            DNxscopeStream((12,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []  # no data yet

        din0 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]

        # th1 is now triggered

        din1 = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]

        din1 = [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
        ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_chanxtochany_hoffset():
    with global_lock:
        assert len(TriggerHandler._instances) == 0
        assert len(TriggerHandler._wait_for_src) == 0

        # chan0 - always off
        dtc0 = DTriggerConfig(ETriggerType.ALWAYS_OFF)
        th0 = TriggerHandler(0, dtc0)

        # chan1 - trigger on chan0 rising endge 4
        hoffset = 2
        level = 4
        srcchan = 0
        dtc1 = DTriggerConfig(
            ETriggerType.EDGE_RISING,
            srcchan=srcchan,
            hoffset=hoffset,
            level=level,
        )
        th1 = TriggerHandler(1, dtc1)

        assert len(th0._instances) == 2
        assert len(th1._instances) == 2

        din0 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []

        din0 = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((-1,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []

        din0 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((4,), ()),
            DNxscopeStream((3,), ()),
            DNxscopeStream((2,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []

        # th1 triggerd with hoffset - but th0 is always off
        din0 = [
            DNxscopeStream((3,), ()),
            DNxscopeStream((4,), ()),
            DNxscopeStream((5,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((10,), ()),
            DNxscopeStream((11,), ()),
            DNxscopeStream((12,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == []  # no data yet

        din0 = [
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
            DNxscopeStream((0,), ()),
        ]
        dout0 = th0.data_triggered(din0)
        assert dout0 == []

        din1 = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]

        # th1 is now triggered
        din1 = [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == [
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
            DNxscopeStream((1,), ()),
        ]

        din1 = [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
        ]
        dout1 = th1.data_triggered(din1)
        assert dout1 == [
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
            DNxscopeStream((-1,), ()),
        ]

        # clean up
        TriggerHandler.cls_cleanup()


def test_triggerhandle_edgerising_hoffset():
    # TODO
    pass


def test_triggerhandle_chanxtochanx():
    pass


def test_triggerhandle_allthesame():
    pass



import pytest  # type: ignore
from nxslib.intf.dummy import DummyDev
from nxslib.nxscope import NxscopeHandler
from nxslib.proto.parse import Parser

from nxscli.iplugin import (
    DPluginDescription,
    EPluginType,
    IPlugin,
    IPluginPlotDynamic,
    IPluginPlotStatic,
    IPluginText,
)
from nxscli.phandler import PluginHandler
from nxscli.trigger import DTriggerConfigReq


class MockPlugin1(IPlugin):
    def __init__(self):
        super().__init__(EPluginType.TEXT)
        self._wait_for = False

    @property
    def stream(self) -> bool:
        return False

    def wait_for_plugin(self) -> bool:
        ret = self._wait_for
        self._wait_for = True
        return ret

    def stop(self) -> None:
        pass

    def data_wait(self, timeout=None) -> bool:
        return True

    def start(self, kwargs) -> bool:
        return True

    def result(self):
        return "1"


class MockPlugin2(IPluginText):
    def __init__(self):
        super().__init__()

    @property
    def stream(self) -> bool:
        return True

    def stop(self) -> None:
        pass

    def data_wait(self, timeout=None) -> bool:
        return True

    def start(self, kwargs) -> bool:
        return True

    def result(self):
        return "2"


class MockPlugin3(IPluginPlotDynamic):
    def __init__(self):
        super().__init__()

    @property
    def stream(self) -> bool:
        return True

    def stop(self) -> None:
        pass

    def data_wait(self, timeout=None) -> bool:
        return True

    def start(self, kwargs) -> bool:
        return True

    def result(self):
        return None


class MockPlugin4(IPluginPlotStatic):
    def __init__(self):
        super().__init__()

    @property
    def stream(self) -> bool:
        return True

    def stop(self) -> None:
        pass

    def data_wait(self, timeout=None) -> bool:
        return False

    def start(self, kwargs) -> bool:
        return True

    def result(self):  # pragma: no cover
        return None


def test_phandler_init():
    # no plugins at the beginning
    p = PluginHandler([])
    assert isinstance(p, PluginHandler)
    assert p.names == []
    p.cleanup()

    # valid data
    plugins = [
        DPluginDescription("plugin1", MockPlugin1),
        DPluginDescription("plugin2", MockPlugin2),
    ]
    p = PluginHandler(plugins=plugins)
    assert isinstance(p, PluginHandler)
    assert p.names == ["plugin1", "plugin2"]
    assert p.plugin_get("plugin1") == MockPlugin1
    assert p.plugin_get("plugin2") == MockPlugin2
    assert p.names == ["plugin1", "plugin2"]

    # add plugin = valid data
    p.plugin_add(("plugin3", MockPlugin3))
    assert p.names == ["plugin1", "plugin2", "plugin3"]

    # plugins instances
    assert isinstance(p.plugins["plugin1"](), MockPlugin1)
    assert isinstance(p.plugins["plugin2"](), MockPlugin2)
    assert isinstance(p.plugins["plugin3"](), MockPlugin3)

    # clean up
    p.cleanup()


@pytest.fixture
def nxscope():
    intf = DummyDev()
    parse = Parser()
    nxscope = NxscopeHandler(intf, parse)
    return nxscope


def test_phandler_connect(nxscope):
    plugins = [
        DPluginDescription("plugin1", MockPlugin1),
        DPluginDescription("plugin2", MockPlugin2),
    ]
    p = PluginHandler(plugins=plugins)

    # nxs not connected
    with pytest.raises(AssertionError):
        _ = p.dev
    with pytest.raises(AssertionError):
        _ = p.stream_start()
    with pytest.raises(AssertionError):
        _ = p.stream_stop()
    with pytest.raises(AssertionError):
        _ = p.channels_configure([])

    # connect nxslib instance
    p.nxscope_connect(nxscope)

    # nxscope should be connected
    assert p.dev is not None

    # chanlist
    p.channels_configure([])
    p.channels_configure([-1], 1)
    p.channels_configure([1, 2], [1, 2], writenow=True)

    # clean up
    p.cleanup()


def test_phandler_enable():
    plugins = [
        DPluginDescription("plugin1", MockPlugin1),
        DPluginDescription("plugin2", MockPlugin2),
    ]
    p = PluginHandler(plugins=plugins)

    # no plugins enabled at default
    assert len(p.enabled) == 0

    pid1 = p.enable("plugin2", **{})
    assert len(p.enabled) == 1
    assert pid1 == 0
    assert p.enabled[0][0] == 0
    assert p.enabled[0][1] == MockPlugin2
    assert p.enabled[0][2] == {}

    pid2 = p.enable("plugin1", **{"arg1": 1})
    assert len(p.enabled) == 2
    assert pid2 == 1
    assert p.enabled[0][0] == 0
    assert p.enabled[0][1] == MockPlugin2
    assert p.enabled[0][2] == {}

    assert p.enabled[0][2] == {}
    assert p.enabled[1][0] == 1
    assert p.enabled[1][1] == MockPlugin1
    assert p.enabled[1][2]["arg1"] == 1

    # we can enable plugins multiple times
    pid3 = p.enable("plugin1", **{"arg1": "test"})
    assert len(p.enabled) == 3
    assert pid3 == 2
    assert p.enabled[0][0] == 0
    assert p.enabled[0][1] == MockPlugin2
    assert p.enabled[0][2] == {}

    assert p.enabled[1][0] == 1
    assert p.enabled[1][1] == MockPlugin1
    assert p.enabled[1][2]["arg1"] == 1

    assert p.enabled[2][0] == 2
    assert p.enabled[2][1] == MockPlugin1
    assert p.enabled[2][2]["arg1"] == "test"

    # disable plugin1
    p.disable(pid1)
    assert len(p.enabled) == 2
    assert p.enabled[0][0] == 1
    assert p.enabled[1][0] == 2

    # once again disable plugin1
    with pytest.raises(AttributeError):
        p.disable(pid1)

    # disable plugin2
    p.disable(pid2)
    assert len(p.enabled) == 1
    assert p.enabled[0][0] == 2

    # once again disable plugin2
    with pytest.raises(AttributeError):
        p.disable(pid2)

    # disable plugin3
    p.disable(pid3)
    assert len(p.enabled) == 0

    # once again disable plugin3
    with pytest.raises(AttributeError):
        p.disable(pid3)

    # clean up
    p.cleanup()


def test_phandler_start_ready(nxscope):
    plugins = [
        DPluginDescription("plugin1", MockPlugin1),
        DPluginDescription("plugin2", MockPlugin2),
        DPluginDescription("plugin3", MockPlugin3),
    ]
    p = PluginHandler(plugins=plugins)
    p.nxscope_connect(nxscope)

    # enable all plugins
    p.enable("plugin1", **{})
    p.enable("plugin2", **{})
    p.enable("plugin3", **{})
    assert len(p.enabled) == 3

    # start plugins
    p.start()

    ret = p.ready()
    assert ret[0].result() == "1"
    assert ret[1].result() == "2"
    assert ret[2].result() is None

    # plugins not need to wait
    assert p.wait_for_plugins() is None

    # stop plugins
    p.stop()

    # clean up
    p.cleanup()


def test_phandler_start_poll(nxscope):
    plugins = [
        DPluginDescription("plugin1", MockPlugin1),
        DPluginDescription("plugin2", MockPlugin2),
        DPluginDescription("plugin3", MockPlugin3),
    ]
    p = PluginHandler(plugins=plugins)
    p.nxscope_connect(nxscope)

    # enable all plugins
    p.enable("plugin1", **{})
    p.enable("plugin2", **{})
    p.enable("plugin3", **{})
    assert len(p.enabled) == 3

    # start plugins
    p.start()

    # poll
    ret = p.poll()
    assert ret[0].result() == "1"
    assert ret[1].result() == "2"
    assert ret[2].result() is None

    # poll once again but all handled
    ret = p.poll()
    assert ret is None

    # plugins not need to wait
    assert p.wait_for_plugins() is None

    # stop plugins
    p.stop()

    # clean up
    p.cleanup()


def test_phandler_start_nostream(nxscope):
    plugins = [DPluginDescription("plugin1", MockPlugin1)]
    p = PluginHandler(plugins=plugins)
    p.nxscope_connect(nxscope)

    # enable all plugins
    p.enable("plugin1", **{})
    assert len(p.enabled) == 1

    # start plugins
    p.start()

    ret = p.ready()
    assert ret[0].result() == "1"

    # plugins not need to wait
    assert p.wait_for_plugins() is None

    # stop plugins
    p.stop()

    # clean up
    p.cleanup()


def test_phandler_start_noready(nxscope):
    plugins = [DPluginDescription("plugin4", MockPlugin4)]
    p = PluginHandler(plugins=plugins)
    p.nxscope_connect(nxscope)

    # enable all plugins
    p.enable("plugin4", **{})
    assert len(p.enabled) == 1

    # start plugins
    p.start()

    # always not ready
    ret = p.ready()
    assert ret == []

    # always not ready
    ret = p.poll()
    assert ret == []

    # plugins not need to wait
    assert p.wait_for_plugins() is None

    # stop plugins
    p.stop()

    # clean up
    p.cleanup()


def test_phandler_trigger():
    p = PluginHandler()

    # default all on
    dt = p.trigger_get(0)
    assert dt.ttype == "on"
    assert dt.srcchan is None
    assert dt.params is None
    dt = p.trigger_get(1)
    assert dt.ttype == "on"
    assert dt.srcchan is None
    assert dt.params is None

    trg = {-1: DTriggerConfigReq("off", None)}
    p.triggers_configure(trg)
    dt = p.trigger_get(0)
    assert dt.ttype == "off"
    assert dt.srcchan is None
    assert dt.params is None
    dt = p.trigger_get(1)
    assert dt.ttype == "off"
    assert dt.srcchan is None
    assert dt.params is None

    trg = {-1: DTriggerConfigReq("on", None)}
    p.triggers_configure(trg)
    dt = p.trigger_get(0)
    assert dt.ttype == "on"
    assert dt.srcchan is None
    assert dt.params is None
    dt = p.trigger_get(1)
    assert dt.ttype == "on"
    assert dt.srcchan is None
    assert dt.params is None

    trg = {0: DTriggerConfigReq("on", None), 1: DTriggerConfigReq("off", None)}
    p.triggers_configure(trg)
    dt = p.trigger_get(0)
    assert dt.ttype == "on"
    assert dt.srcchan is None
    assert dt.params is None
    dt = p.trigger_get(1)
    assert dt.ttype == "off"
    assert dt.srcchan is None
    assert dt.params is None

    # clean up
    p.cleanup()



"""Module containint the trigger configuration commnad for CLI."""

from typing import TYPE_CHECKING

import click

from nxscli.cli.environment import Environment, pass_environment
from nxscli.cli.types import Trigger

if TYPE_CHECKING:
    from nxscli.trigger import DTriggerConfigReq


###############################################################################
# Command: cmd_trig
###############################################################################


@click.command(name="trig")
@click.argument("triggers", type=Trigger())
@pass_environment
def cmd_trig(
    ctx: Environment, triggers: dict[int, "DTriggerConfigReq"]
) -> bool:
    """[config] Triggers configuration.

    This command configure software triggers.

    Format: '[channel]:[trigger][#chan_source][@chan_vector][parameters]'

    You can define many triggers at once, just separate them with
    a semicolon (;). The configuration request must be surrounded
    by quotation marks. All spaces are ignored.

    If channel is equal to character 'g' then the trigger is considered
    global for all channels.

    You can precisesly configure the triggers for a given plugin using
    the '--trig' option.

    \b
    Supported 'trigger' options:
       on - always on
       off - always off
       er - edge rising, parameters: [hoffset, level]
       ef - edge falling: [hoffset, level]

    \b
    where:
       hoffset - horizontal offset for triggered data
       level - trigger level

    Default: all channels on ('g:on').

    \b
    Some triggers configuration examples:
    - all chans triggered on a rising edge, hoffset=0, level=10.0:
      'g:er,0,10.0'
    - all chans triggered on a chan 0 rising edge, hoffset=0, level=10.0:
      'g:er#0,0,10.0'
    - chan 1 triggered on a vect 3 rising edge, hoffset=0, level=10.0:
      '1:er@3,0,10.0'
    - chan 2 triggered on chan 3 falling edge, hoffset=10, level=-10:
      '2:er#2,10,-10'
    - chan 2 triggered on chan 3 vect 1 falling edge, hoffset=0, level=1,
      chan 3 always on,
      chan 4 always off,
      white spaces ignored:
      '2 : er#2@1, 0, 1; 3 : on; 4 : off'
    """  # noqa: D301
    assert ctx.phandler
    ctx.triggers = triggers
    # configure triggers
    ctx.phandler.triggers_configure(ctx.triggers)

    return True



"""Module containint the CLI logic for Nxslib."""

import sys
from typing import Any

import click
from nxslib.proto.parse import Parser

from nxscli.cli.environment import Environment, pass_environment
from nxscli.iplugin import EPluginType, IPlugin
from nxscli.logger import logger
from nxscli.phandler import PluginHandler
from nxscli.plugins_loader import commands_list, interfaces_list, plugins_list

###############################################################################
# Function: main
###############################################################################


@click.group()
@click.option(
    "--debug/--no-debug",
    default=False,
    is_flag=True,
    envvar="NXSCLI_DEBUG",
)
@pass_environment
def main(ctx: Environment, debug: bool) -> bool:
    """Nxscli - Command-line clinet to the NxScope."""
    ctx.debug = debug
    if debug:  # pragma: no cover
        logger.setLevel("DEBUG")
    else:
        logger.setLevel("INFO")

    ctx.phandler = PluginHandler(plugins_list)
    parse = Parser()
    ctx.parser = parse
    ctx.triggers = {}

    click.get_current_context().call_on_close(cli_on_close)

    return True


###############################################################################
# Function: handle_plugin
###############################################################################


def handle_plugin(plugin: IPlugin) -> tuple[EPluginType, Any] | None:
    """Handle a given plugin."""
    if plugin.ptype == EPluginType.TEXT:
        # print output
        print(plugin.result())
        return None

    elif plugin.ptype == EPluginType.STATIC:  # pragma: no cover
        plot = plugin.result()
        return (EPluginType.STATIC, plot)

    elif plugin.ptype == EPluginType.ANIMATION:  # pragma: no cover
        plot = plugin.result()
        return (EPluginType.ANIMATION, plot)

    elif plugin.ptype == EPluginType.FILE:
        return None

    elif plugin.ptype == EPluginType.NONE:
        return None

    else:
        raise AssertionError


###############################################################################
# Function: plugin_loop
###############################################################################


def plugin_loop(ctx: Environment) -> list[Any]:
    """Plugin loop."""
    assert ctx.phandler
    ret: list[Any] = []
    while True:
        plugins = ctx.phandler.poll()
        if plugins is None:
            break

        if len(plugins) > 0:
            for x in plugins:
                r = handle_plugin(x)
                if ret:  # pragma: no cover
                    ret.append(r)
        else:  # pragma: no cover
            pass

    return ret


###############################################################################
# Function: wait_for_plugins
###############################################################################


def wait_for_plugins(ctx: Environment) -> None:  # pragma: no cover
    """Wait for plugins."""
    assert ctx.phandler
    ctx.phandler.wait_for_plugins()


###############################################################################
# Function: cli_on_close
###############################################################################


@pass_environment
def cli_on_close(ctx: Environment) -> bool:
    """Handle requested plugins on click close."""
    assert ctx.phandler
    # do not show any errors if it was help request
    if "--help" in sys.argv:  # pragma: no cover
        ctx.phandler.cleanup()
        return True

    if ctx.interface is False:
        ctx.phandler.cleanup()
        return False

    if len(ctx.phandler.enabled) == 0:
        click.secho("ERROR: No plugins selected !", err=True, fg="red")
        ctx.phandler.cleanup()
        return False

    if ctx.needchannels:
        if ctx.channels is None:  # pragma: no cover
            click.secho("ERROR: No channels selected !", err=True, fg="red")
            ctx.phandler.cleanup()
            return False

    # connect nxscope to phandler
    assert ctx.nxscope
    ctx.phandler.nxscope_connect(ctx.nxscope)

    # configure channles after connected to nxscope
    if ctx.needchannels and ctx.channels:
        ctx.phandler.channels_configure(ctx.channels[0], ctx.channels[1])

    # start plugins
    ctx.phandler.start()

    if ctx.waitenter:  # pragma: no cover
        _ = input("wait for Enter key...")

    # plugins loop
    plugin_loop(ctx)

    # wait for plugin
    wait_for_plugins(ctx)

    print("closing...")
    ctx.phandler.stop()
    ctx.phandler.nxscope_disconnect()
    ctx.phandler.cleanup()

    return True


###############################################################################
# Function: click_final_init
###############################################################################


def click_final_init() -> None:
    """Handle final Click initialization."""
    # add interfaces
    for intf in interfaces_list:
        main.add_command(intf)

    # add commands to interfaces
    for group in interfaces_list:
        for cmd in commands_list:
            group.add_command(cmd)


# final click initialization
click_final_init()



"""Module containing Nxscli stream data trigger logic."""

import itertools
import weakref
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from threading import Lock
from typing import TYPE_CHECKING, Any, Iterator

from nxscli.logger import logger

if TYPE_CHECKING:
    from nxslib.nxscope import DNxscopeStream

###############################################################################
# Enum: ETriggerType
###############################################################################


class ETriggerType(Enum):
    """Trigger types."""

    ALWAYS_OFF = 0
    ALWAYS_ON = 1
    EDGE_RISING = 2
    EDGE_FALLING = 3


###############################################################################
# Enum: DTriggerState
###############################################################################


@dataclass
class DTriggerState:
    """The class representing trigger state."""

    state: bool
    idx: int


###############################################################################
# Class: DDTriggerConfigReq
###############################################################################


@dataclass
class DTriggerConfigReq:
    """The class representing trigger configuration request."""

    ttype: str
    srcchan: int | None
    vect: int = 0
    params: list[Any] | None = None


###############################################################################
# Function: trigger_from_req
###############################################################################


def trigger_from_req(req: DTriggerConfigReq) -> "DTriggerConfig":
    """Get trigger configuration from string.

    :param req: trigger configuration request
    """
    if req.ttype == "off":
        # no arguments
        dtc = DTriggerConfig(ETriggerType.ALWAYS_OFF)
    elif req.ttype == "on":
        # no arguments
        dtc = DTriggerConfig(ETriggerType.ALWAYS_ON)
    elif req.ttype == "er":
        # argument 1 horisontal offset
        # argument 2 trigger level
        assert req.params
        hoffset = int(req.params[0])
        level = float(req.params[1])
        dtc = DTriggerConfig(
            ETriggerType.EDGE_RISING, req.srcchan, req.vect, hoffset, level
        )
    elif req.ttype == "ef":
        # argument 1 horisontal offset
        # argument 2 trigger level
        assert req.params
        hoffset = int(req.params[0])
        level = float(req.params[1])
        dtc = DTriggerConfig(
            ETriggerType.EDGE_FALLING, req.srcchan, req.vect, hoffset, level
        )
    else:
        raise AssertionError
    return dtc


###############################################################################
# Class: DTriggerConfig
###############################################################################


@dataclass
class DTriggerConfig:
    """Trigger configuration."""

    ttype: ETriggerType
    srcchan: int | None = None
    vect: int = 0
    hoffset: int = 0
    level: float | None = None


###############################################################################
# Class: TriggerHandler
###############################################################################


class TriggerHandler(object):
    """The class used to handler stream data trigger."""

    _instances: weakref.WeakSet["TriggerHandler"] = weakref.WeakSet()
    _wait_for_src: weakref.WeakSet["TriggerHandler"] = weakref.WeakSet()

    def __new__(cls, chan: int, config: DTriggerConfig) -> "TriggerHandler":
        """Create a new instance and store reference in a weak set."""
        instance = object.__new__(cls)
        cls._instances.add(instance)
        # create additional flag that helps during instance cleanup
        object.__setattr__(instance, "_initdone", False)
        return instance

    def __init__(self, chan: int, config: DTriggerConfig) -> None:
        """Initialize a stream data trigger handler."""
        self._config = config
        self._cache: list["DNxscopeStream"] = []
        self._chan: int = chan
        self._trigger: DTriggerState = DTriggerState(False, 0)
        self._triger_done = False

        # trigger source channel reference
        self._src: "TriggerHandler" | None = None
        # connected cross channels
        self._cross: list["TriggerHandler"] = []
        self._src_configured = False
        # configure cross trigger
        self._src_configured = True
        self._config_crosschan()

        # check pending cross channels
        self._pending_crosschan()

        # cross trigger status protected with lock
        self._lock = Lock()
        self._cross_trigger: DTriggerState = DTriggerState(False, 0)
        self._initdone = True

    def __del__(self) -> None:
        """Clean up."""
        self.cleanup()

    def _config_crosschan(self) -> None:
        # subscribe to source channel if needed
        src = None
        if self._config.srcchan is not None:
            # get the first available trigger related to srcchan
            for x in TriggerHandler._instances:
                if x.chan == self._config.srcchan:
                    src = x
                    break

            if not src:
                # not found source channel
                # trigger not configured yet
                self._src_configured = False
                # wait for source channel
                TriggerHandler._wait_for_src.add(self)
            else:
                # connect source
                self.source_set(src)
                # subscribe to source
                src.subscribe_cross(self)

    def _pending_crosschan(self) -> None:
        # check if any instance wait for us
        # NOTE: there can be many waiting instances
        tmp = []
        for inst in TriggerHandler._wait_for_src:
            if inst.config.srcchan == self.chan:
                # subscribe us and set source
                self.subscribe_cross(inst)
                inst.source_set(self)
                tmp.append(inst)
        # remove all handled instnaces from set
        for inst in tmp:
            TriggerHandler._wait_for_src.remove(inst)

    def _pairwise(
        self, iterable: list["DNxscopeStream"]
    ) -> Iterator[tuple["DNxscopeStream", "DNxscopeStream"]]:
        (a, b) = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)

    def _alwaysoff(self, _: list["DNxscopeStream"]) -> DTriggerState:
        # reset cache
        self._cache = []
        return DTriggerState(False, 0)

    def _alwayson(self, _: list["DNxscopeStream"]) -> DTriggerState:
        return DTriggerState(True, 0)

    def _edgerising(
        self, combined: list["DNxscopeStream"], vect: int, level: float
    ) -> DTriggerState:
        ret = False
        idx = 0
        for a, b in self._pairwise(combined):
            if a.data[vect] <= level < b.data[vect]:
                ret = True
                idx = idx
                break
            idx += 1

        if not ret:
            idx = 0

        return DTriggerState(ret, idx)

    def _edgefalling(
        self, combined: list["DNxscopeStream"], vect: int, level: float
    ) -> DTriggerState:
        ret = False
        idx = 0
        for a, b in self._pairwise(combined):
            if a.data[vect] >= level > b.data[vect]:
                ret = True
                idx = idx
                break
            idx += 1

        if not ret:
            idx = 0

        return DTriggerState(ret, idx)

    def _is_self_trigger(
        self, combined: list["DNxscopeStream"], config: DTriggerConfig
    ) -> DTriggerState:
        if config.ttype is ETriggerType.ALWAYS_OFF:
            return self._alwaysoff(combined)
        elif config.ttype is ETriggerType.ALWAYS_ON:
            return self._alwayson(combined)
        elif config.ttype is ETriggerType.EDGE_RISING:
            assert config.level is not None
            return self._edgerising(combined, config.vect, config.level)
        elif config.ttype is ETriggerType.EDGE_FALLING:
            assert config.level is not None
            return self._edgefalling(combined, config.vect, config.level)
        else:
            raise AssertionError

    def _is_triggered(self, combined: list["DNxscopeStream"]) -> DTriggerState:
        if self._trigger.state:
            # make sure that idx is 0
            return DTriggerState(True, 0)

        # cross-channel trigger
        if self._config.srcchan is not None:
            # source must be connected
            if not self._src_configured:
                raise AssertionError
            return self.cross_trigger

        # self-triggered
        return self._is_self_trigger(combined, self._config)

    def _cross_channel_handle(self, combined: list["DNxscopeStream"]) -> None:
        for cross in self._cross:
            if cross.cross_trigger.state is True:
                continue
            # check cross channel requirements for trigger
            trigger = self._is_self_trigger(combined, cross.config)
            # signal if triggered
            if trigger.state is True:
                cross.cross_trigger = trigger

    @property
    def cross_trigger(self) -> DTriggerState:
        """Get cross tirgger state."""
        with self._lock:
            ret = deepcopy(self._cross_trigger)
        return ret

    @cross_trigger.setter
    def cross_trigger(self, val: DTriggerState) -> None:
        """Set cross tirgger state.

        :param val: cross triger state
        """
        with self._lock:
            self._cross_trigger = val

    @property
    def chan(self) -> int:
        """Get channel id associated with this trigger."""
        return self._chan

    @property
    def config(self) -> DTriggerConfig:
        """Get trigger configuration."""
        return self._config

    @classmethod
    def cls_cleanup(cls: type["TriggerHandler"]) -> None:
        """Clean up all instances."""
        for x in cls._instances:
            x.cleanup(remove_self=False)
        # clear set
        cls._instances.clear()

    def cleanup(self, remove_self: bool = True) -> None:
        """Clean up instance."""
        if self._initdone:
            if self._src:
                self._src.unsubscribe_cross(self)
            if self in TriggerHandler._wait_for_src:
                TriggerHandler._wait_for_src.remove(self)
        else:  # pragma: no cover
            pass
        if remove_self:
            if self in self._instances:
                self._instances.remove(self)  # pragma: no cover

    def source_set(self, inst: "TriggerHandler") -> None:
        """Set source instance."""
        assert inst
        logger.info("set source trigger = %s for %s", str(inst), str(self))
        self._src = inst
        self._src_configured = True

    def subscribe_cross(self, inst: "TriggerHandler") -> None:
        """Subscribe as cross-channel trigger.

        :param inst: trigger handler instance to subscribe
        """
        self._cross.append(inst)

    def unsubscribe_cross(self, inst: "TriggerHandler") -> None:
        """Unsubscribe as cross-channel trigger.

        :param inst: trigger handler instance to unsubscribe
        """
        for i, cross in enumerate(self._cross):
            if cross is inst:  # pragma: no cover
                self._cross.pop(i)

    def data_triggered(
        self, data: list["DNxscopeStream"]
    ) -> list["DNxscopeStream"]:
        """Get triggered data.

        :param data: stream data
        """
        combined = self._cache + data

        self._trigger = self._is_triggered(combined)

        # check all cross-channel triggers
        self._cross_channel_handle(combined)

        if not self._trigger.state:
            # not triggered yet
            ret = []
            # update cache
            clen = len(self._cache)
            self._cache = combined[clen - self._config.hoffset :]
        else:
            # one time hoffset for trigger
            if not self._triger_done:
                hoffset = self._config.hoffset
                self._triger_done = True
            else:
                hoffset = 0

            # return data with a configured horisontal offset
            ret = combined[self._trigger.idx - hoffset :]
            # reset cache
            self._cache = []

        return ret



"""Module containing CSV plugin."""

import csv
from typing import TYPE_CHECKING, Any

from nxscli.idata import PluginData, PluginQueueData
from nxscli.iplugin import IPluginFile
from nxscli.logger import logger
from nxscli.pluginthr import PluginThread

if TYPE_CHECKING:
    from nxslib.nxscope import DNxscopeStream


###############################################################################
# Class: PluginCsv
###############################################################################


class PluginCsv(PluginThread, IPluginFile):
    """Plugin that capture data to CSV files."""

    def __init__(self) -> None:
        """Initialize a CSV plugin."""
        IPluginFile.__init__(self)
        PluginThread.__init__(self)

        self._data: "PluginData"
        self._path: str
        self._meta_string = False
        self._csvwriters: list[Any] = []

    def _csvwriters_open(self) -> list[Any]:
        csvwriters = []
        for pdata in self._data.qdlist:
            chanpath = self._path + "_chan" + str(pdata.chan) + ".csv"
            csvfile = open(chanpath, "w", newline="")
            csvwriter = csv.writer(
                csvfile,
                delimiter=" ",
                quotechar="|",
                escapechar="\\",
                quoting=csv.QUOTE_MINIMAL,
            )

            csvwriters.append([csvwriter, csvfile])

        return csvwriters

    def _sample_row_get(self, sample: "DNxscopeStream") -> tuple[Any, Any]:
        # covert to string
        if self._meta_string:
            return (sample.data, bytes(list(sample.meta)).decode())
        else:
            return sample.data, sample.meta

    def _init(self) -> None:
        assert self._phandler
        # open writers
        self._csvwriters = self._csvwriters_open()

    def _final(self) -> None:
        # close all files
        for csvwriter in self._csvwriters:
            csvwriter[1].close()

        logger.info("csv capture DONE")

    def _handle_samples(
        self, data: list["DNxscopeStream"], pdata: "PluginQueueData", j: int
    ) -> None:
        # store data
        for sample in data:
            if not self._nostop:  # pragma: no cover
                # ignore data if capture done for channel
                if self._datalen[j] >= self._samples:
                    break

            # write row
            self._csvwriters[j][0].writerow(self._sample_row_get(sample))

            # one sample
            self._datalen[j] += 1

    def start(self, kwargs: Any) -> bool:
        """Start CSV plugin.

        :param kwargs: implementation specific arguments
        """
        assert self._phandler

        logger.info("start csv %s", str(kwargs))

        self._samples = kwargs["samples"]
        self._path = kwargs["path"]
        self._meta_string = kwargs["metastr"]
        self._nostop = kwargs["nostop"]

        chanlist = self._phandler.chanlist_plugin(kwargs["channels"])
        trig = self._phandler.triggers_plugin(chanlist, kwargs["trig"])

        cb = self._phandler.cb_get()
        self._data = PluginData(chanlist, trig, cb)

        if not self._data.qdlist:  # pragma: no cover
            return False

        self.thread_start(self._data)

        return True

    def result(self) -> None:
        """Get CSV plugin result."""
        # TODO: return file handler ?
        return  # pragma: no cover



"""Module containint the channels configuration command for CLI."""

from typing import Any

import click

from nxscli.cli.environment import Environment, pass_environment
from nxscli.cli.types import Channels, Divider, divider_option_help

###############################################################################
# Command: cmd_chan
###############################################################################


@click.command(name="chan")
@click.argument("channels", required=True, type=Channels())
@click.option(
    "--divider", default="0", type=Divider(), help=divider_option_help
)
@pass_environment
def cmd_chan(ctx: Environment, channels: list[int], divider: Any) -> bool:
    """[config] Channels declaration and configuration.

    This command configure and enable given channels.
    The channels must be integer separated by commas, eg 'chan 1,2,3'.
    All availalbe channels can be configured with a single word 'all'.

    By default all channels from this command are passed to all plugins.

    \b
    You can precisesly configure the channels for a given plugin using
    the '--chan' option.
    """  # noqa: D301
    ctx.channels = (channels, divider)

    return True



"""Module containing dummy capture plugin command."""

from typing import TYPE_CHECKING

import click

from nxscli.cli.environment import Environment, pass_environment
from nxscli.cli.types import Samples, capture_options

if TYPE_CHECKING:
    from nxscli.trigger import DTriggerConfigReq


###############################################################################
# Command: cmd_pnone
###############################################################################


@click.command(name="pnone")
@click.argument("samples", type=Samples(), required=True)
@capture_options
@pass_environment
def cmd_pnone(
    ctx: Environment,
    samples: int,
    chan: list[int],
    trig: dict[int, "DTriggerConfigReq"],
) -> bool:
    """[plugin] Capture data and do nothing with them.

    If SAMPLES argument is set to 'i' then we capture data until enter
    is press.
    """  # noqa: D301
    # wait for enter if samples set to 'i'
    assert ctx.phandler
    if samples == 0:  # pragma: no cover
        ctx.waitenter = True

    ctx.phandler.enable(
        "none",
        samples=samples,
        channels=chan,
        trig=trig,
        nostop=ctx.waitenter,
    )

    ctx.needchannels = True

    return True



"""Module containing Nxscli plugin thread common logic."""

import threading
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from nxslib.thread import ThreadCommon

if TYPE_CHECKING:
    from nxslib.nxscope import DNxscopeStream

    from nxscli.idata import PluginData, PluginQueueData


###############################################################################
# Class: PluginThread
###############################################################################


class PluginThread(ABC):
    """The Nxscli plugin thread."""

    def __init__(self) -> None:
        """Initialize a Nxslib plugin thread."""
        self._thread = ThreadCommon(
            self._thread_common, self._init_common, self._final_common
        )

        self._samples: int
        self._nostop = False
        self._ready = threading.Event()
        self._datalen: list[int] = []
        self._plugindata: "PluginData"

    def _thread_common(self) -> None:
        assert self._plugindata
        # get samples
        for j, pdata in enumerate(self._plugindata.qdlist):
            # get data from queue
            data = pdata.queue_get(block=True, timeout=1)

            if not self._nostop:  # pragma: no cover
                # ignore data if capture done for channel
                if self._datalen[j] >= self._samples:
                    continue

            # handle samples
            self._handle_samples(data, pdata, j)

        # break loop if done
        if self._is_done(self._datalen):
            self._thread.stop_set()

    def _init_common(self) -> None:
        self._init()
        self._datalen = [0 for _ in range(len(self._plugindata.qdlist))]

    def _final_common(self) -> None:
        self._final()
        self._ready.set()

    @property
    def stream(self) -> bool:
        """Return True if this plugin needs stream."""
        return True

    def stop(self) -> None:
        """Stop capture plugin."""
        self._thread.thread_stop()

    def data_wait(self, timeout: float = 0.0) -> bool:
        """Return True if data are ready.

        :param timeout: data wait timeout
        """
        if self._nostop:  # pragma: no cover
            return True
        return self._ready.wait(timeout)

    def thread_start(self, plugindata: "PluginData") -> None:
        """Start working thread."""
        self._plugindata = plugindata
        self._thread.thread_start()

    def _is_done(self, datalen: list[int]) -> bool:
        if not self._nostop:
            # check if capture done
            done = True
            for x in datalen:
                if x < self._samples:
                    done = False
        else:  # pragma: no cover
            done = False
        return done

    @abstractmethod
    def _handle_samples(
        self, data: list["DNxscopeStream"], pdata: "PluginQueueData", j: int
    ) -> None:
        """Handle samples from queue and update datalen."""

    @abstractmethod
    def _init(self) -> None:
        """Logic called before thread loop."""

    @abstractmethod
    def _final(self) -> None:
        """Logic called after thread loop."""



import nxscli
import nxscli.ext_commands
import nxscli.ext_interfaces
import nxscli.ext_plugins


def test_extplugins():
    assert nxscli.__version__

    assert isinstance(nxscli.ext_plugins.plugins_list, list)
    assert isinstance(nxscli.ext_commands.commands_list, list)
    assert isinstance(nxscli.ext_interfaces.interfaces_list, list)



"""Module containing devinfo plugin command."""

import click

from nxscli.cli.environment import Environment, pass_environment

###############################################################################
# Command: cmd_pdevinfo
###############################################################################


@click.command(name="pdevinfo")
@pass_environment
def cmd_pdevinfo(ctx: Environment) -> bool:
    """[plugin] Show information about the connected NxScope device."""
    assert ctx.phandler
    ctx.phandler.enable("devinfo")

    return True



"""Plugins loader."""

from importlib.metadata import entry_points
from typing import TYPE_CHECKING

import nxscli.ext_commands
import nxscli.ext_interfaces
import nxscli.ext_plugins
from nxscli.logger import logger

if TYPE_CHECKING:
    import click

    from nxscli.iplugin import DPluginDescription

plugins_list: list["DPluginDescription"] = []
commands_list: list["click.Command"] = []
interfaces_list: list["click.Group"] = []

# default plugins
plugins_list.extend(nxscli.ext_plugins.plugins_list)
# default commands
commands_list.extend(nxscli.ext_commands.commands_list)
# default interfaces
interfaces_list.extend(nxscli.ext_interfaces.interfaces_list)

# load external plugins
eps = entry_points(group="nxscli.extensions")
for entry in eps:  # pragma: no cover
    logger.info("loading %s %s ...", entry.name, entry.value)
    plugin = entry.load()
    if entry.name == "plugins":
        plugins_list.extend(plugin.plugins_list)
    elif entry.name == "commands":
        commands_list.extend(plugin.commands_list)
    elif entry.name == "interfaces":
        interfaces_list.extend(plugin.interfaces_list)
    else:
        raise AssertionError("Unsupported entry name")



"""Module containing the Click types."""

from typing import Any

import click

from nxscli.trigger import DTriggerConfigReq

###############################################################################
# Function: get_list_from_str
###############################################################################


def get_list_from_str(value: str, char: str = ",") -> list[str]:
    """Get list of values from string argument."""
    if not len(value):
        return []
    # remove white spaces
    tmp = value.replace(" ", "")
    # one separator
    return tmp.split(char)


###############################################################################
# Function: get_list_from_str2
###############################################################################


def get_list_from_str2(
    value: str, char1: str = ",", char2: str = ";"
) -> list[list[str]]:
    """Get list of values from string argument."""
    if not len(value):
        return []
    # remove white spaces
    tmp = value.replace(" ", "")

    # two separators, start from the second one
    ch2 = tmp.split(char2)
    ch1 = []
    for ch in ch2:
        ch1.append(ch.split(char1))
    return ch1


###############################################################################
# Class: Channels
###############################################################################


class Channels(click.ParamType):
    """Parse channels argument."""

    name = "channels"

    def convert(self, value: Any, param: Any, ctx: Any) -> list[int]:
        """Convert channels argument."""
        lint = []
        if value == "all":
            # special case to indicate all channels
            lint.append(-1)
            return lint

        lstr = get_list_from_str(value)
        for ch in lstr:
            assert ch.isnumeric(), "channel id must be a valid integer"
            chan = int(ch)
            if chan < 0 or chan > 255:
                raise click.BadParameter(
                    "channel id must be in range [0, 255]"
                )
            lint.append(chan)

        return lint


###############################################################################
# Class: Samples
###############################################################################


class Samples(click.ParamType):
    """Parse samples argument."""

    name = "samples"

    def convert(self, value: Any, param: Any, ctx: Any) -> int:
        """Convert samples argument."""
        if value.isnumeric():
            return int(value)
        elif value == "i":
            return -1
        else:
            raise click.BadParameter("samples must be a valid integer or 'i'")


###############################################################################
# Class: Trigger
###############################################################################


class Trigger(click.ParamType):
    """Parse trigger argument."""

    name = "trigger"

    req_split = ";"
    req_assign = ":"
    req_separator = ","
    req_cross = "#"
    req_global = "g"
    req_vect = "@"

    def convert(
        self, value: Any, param: Any, ctx: Any
    ) -> dict[int, DTriggerConfigReq]:
        """Convert trigger argument."""
        tmp = get_list_from_str(value, self.req_split)
        # get configurations
        ret = {}
        for trg in tmp:
            schan, params = trg.split(self.req_assign)
            tmp = params.split(self.req_separator)

            # decore trigger cross channel and channel vector
            vect_idx = tmp[0].find(self.req_vect)
            cross_idx = tmp[0].find(self.req_cross)
            if vect_idx != -1 and cross_idx != -1:
                if vect_idx > cross_idx:
                    trg = tmp[0][:cross_idx]
                    cross = int(tmp[0][cross_idx + 1 : vect_idx])
                    vect = int(tmp[0][vect_idx + 1 :])
                else:
                    trg = tmp[0][:vect_idx]
                    vect = int(tmp[0][vect_idx + 1 : cross_idx])
                    cross = int(tmp[0][cross_idx + 1 :])
            elif vect_idx == -1 and cross_idx != -1:
                trg, cross_s = tmp[0].split(self.req_cross)
                cross = int(cross_s)
                vect = 0
            elif vect_idx != -1 and cross_idx == -1:
                trg, vect_s = tmp[0].split(self.req_vect)
                vect = int(vect_s)
                cross = None
            else:
                trg = tmp[0]
                vect = 0
                cross = None

            cfg = tmp[1:]
            # special case for global configuration
            if schan == self.req_global:
                chan = -1
            else:
                chan = int(schan)

            # reset cross source if we point to ourself
            if cross == chan:
                cross = None

            req = DTriggerConfigReq(trg, cross, vect, cfg)
            ret[chan] = req
        return ret


###############################################################################
# Class: Divider
###############################################################################


class Divider(click.ParamType):
    """Parse divider argument."""

    name = "divider"

    def convert(self, value: Any, param: Any, ctx: Any) -> list[int] | int:
        """Convert divider argument."""
        lstr = get_list_from_str(value)
        lint = []
        for d in lstr:
            assert d.isnumeric(), "divider must be a valid integer"
            div = int(d)
            if div < 0 or div > 255:
                raise click.BadParameter("divnel id must be in range [0, 255]")
            lint.append(div)

        # return as int if one element
        if len(lint) == 1:
            return lint[0]
        # else return list
        return lint


###############################################################################
# Class: StringList
###############################################################################


class StringList(click.ParamType):
    """Parse a string list argument."""

    name = "stringlist"

    def __init__(self, ch1: str = ",") -> None:
        """Initialize parser."""
        super().__init__()
        self._ch1 = ch1

    def convert(self, value: Any, param: Any, ctx: Any) -> list[str]:
        """Convert a string list argument."""
        return get_list_from_str(value, self._ch1)


###############################################################################
# Class: StringList2
###############################################################################


class StringList2(click.ParamType):  # pragma: no cover
    """Parse a string list argument (2 separators)."""

    name = "stringlist2"

    def __init__(self, ch1: str = ",", ch2: str = ";") -> None:
        """Initialize parser."""
        super().__init__()
        self._ch1 = ch1
        self._ch2 = ch2

    def convert(self, value: Any, param: Any, ctx: Any) -> list[list[str]]:
        """Convert a string list argument."""
        return get_list_from_str2(value, self._ch1, self._ch2)


###############################################################################
# Globals: stirngs
###############################################################################


channels_option_help = """Plugin specific channels configuration,
                           for details look at 'chan' command"""  # noqa: D301
trigger_option_help = """Plugin specific triggers configuration,
                          for details look at 'tirg' command"""  # noqa: D301
divider_option_help = """Configure divider for a given channels.
                          Use a single integer to configure all channels with
                          the same divider value, or use a list of integers
                          (separated by commas) to directly configure the
                          channels."""


###############################################################################
# Decorator: capture_options
###############################################################################


# common capture options
_capture_options = (
    click.option(
        "--chan",
        default=None,
        type=Channels(),
        help=channels_option_help,
    ),
    click.option(
        "--trig",
        default=None,
        type=Trigger(),
        help=trigger_option_help,
    ),
)


def capture_options(fn: Any) -> Any:
    """Decorate command with common capture options decorator."""
    for decorator in reversed(_capture_options):
        fn = decorator(fn)
    return fn



"""Default plugins."""

from nxscli.iplugin import DPluginDescription
from nxscli.plugins.csv import PluginCsv
from nxscli.plugins.devinfo import PluginDevinfo
from nxscli.plugins.none import PluginNone
from nxscli.plugins.printer import PluginPrinter

plugins_list = [
    DPluginDescription("devinfo", PluginDevinfo),
    DPluginDescription("csv", PluginCsv),
    DPluginDescription("none", PluginNone),
    DPluginDescription("printer", PluginPrinter),
]



"""Module containing CSV plugin command."""

from typing import TYPE_CHECKING

import click

from nxscli.cli.environment import Environment, pass_environment
from nxscli.cli.types import Samples, capture_options

if TYPE_CHECKING:
    from nxscli.trigger import DTriggerConfigReq


###############################################################################
# Command: cmd_pcsv
###############################################################################


@click.command(name="pcsv")
@click.argument("samples", type=Samples(), required=True)
@click.argument("path", type=click.Path(resolve_path=False), required=True)
@capture_options
@click.option(
    "--metastr", default=False, is_flag=True, help="store metadata as string"
)
@pass_environment
def cmd_pcsv(
    ctx: Environment,
    samples: int,
    path: str,
    chan: list[int],
    trig: dict[int, "DTriggerConfigReq"],
    metastr: bool,
) -> bool:
    """[plugin] Store samples in CSV files.

    Each configured channel will be stored in a separate file.
    If SAMPLES argument is set to 'i' then we capture data until enter
    is press.
    """  # noqa: D301
    # wait for enter if samples set to 'i'
    assert ctx.phandler
    if samples == 0:  # pragma: no cover
        ctx.waitenter = True

    ctx.phandler.enable(
        "csv",
        samples=samples,
        path=path,
        channels=chan,
        trig=trig,
        metastr=metastr,
        nostop=ctx.waitenter,
    )

    ctx.needchannels = True

    return True
